package main

import "encoding/xml"
import "errors"
import "fmt"
import "golang.org/x/sys/unix"
import "golang.org/x/tools/blog/atom"
import "html"
import "io/ioutil"
import "log"
import "os"
import "regexp"
import "strconv"
import "strings"
import "time"

const bugzillaDateFormat = "2006-01-02 15:04:05 -0700"
const userAgentName = "bugzilla2atom"

// Regex for detection of bug numbers in comments. Used in convertXmlToAtom.
// The pre region is used so we don't recognize escaped html as &#1234; as bug number.
var bugNumberRegex *regexp.Regexp = regexp.MustCompile(`(?P<pre>\A|[^&])(?P<all>(?:[Bb]ug |#)(?P<num>\d+))`)

// Regex for detection of attachment uploads in comments. Used in convertXmlToAtom.
var attachmentUploadRegex *regexp.Regexp = regexp.MustCompile(`(?P<pre>\ACreated )(?P<all>attachment (?P<num>\d+))`)

// Somewhat crude regex for detection of most urls in comments. Used in convertXmlToAtom.
var urlRegex *regexp.Regexp = regexp.MustCompile(`\b(?P<all>\w+://\S+)`)

func min(a int, b int) int {
	if a < b {
		return a
	} else {
		return b
	}
}

// converts the given xml string into an atom feed
func convertXmlToAtom(inXml []byte) (string, error) {
	type Who struct {
		Name     string `xml:",chardata"`
		RealName string `xml:"name,attr"`
	}

	getFormatedName := func(w Who) string {
		if w.RealName == "" {
			return w.Name
		} else {
			return w.RealName + " (" + w.Name + ")"
		}
	}

	type Comment struct {
		CommentId    int    `xml:"commentid"`
		CommentCount *int   `xml:"comment_count"`
		AttachmentID int    `xml:"attachid"`
		Who          Who    `xml:"who"`
		When         string `xml:"bug_when"`
		Text         string `xml:"thetext"`
	}

	type InResult struct {
		Urlbase     string    `xml:"urlbase,attr"`
		BugId       int       `xml:"bug>bug_id"`
		Description string    `xml:"bug>short_desc"`
		Comments    []Comment `xml:"bug>long_desc"`
	}

	inResult := InResult{}
	err := xml.Unmarshal(inXml, &inResult)
	if err != nil {
		return "", errors.New(fmt.Sprintf("Error during unmarshalling the xml: %s", err))
	} else if len(inResult.Comments) == 0 {
		// One comment, the initial one, should always be available
		err := errors.New("Zero comments in bug. There should be at least the initial one.")
		return "", err
	}

	updateTime, err := time.Parse(bugzillaDateFormat, inResult.Comments[len(inResult.Comments)-1].When)
	if err != nil {
		return "", errors.New(fmt.Sprintf("Couldn't parse updateTime in initial comment: %s", err))
	}

	inUrl := fmt.Sprintf("%sshow_bug.cgi?id=%d", inResult.Urlbase, inResult.BugId)
	bugUrl := fmt.Sprintf("%sshow_bug.cgi?id=", inResult.Urlbase)
	attachmentUrl := fmt.Sprintf("%sattachment.cgi?id=", inResult.Urlbase)

	feed := &atom.Feed{
		Title:   inResult.Description,
		ID:      inUrl,
		Link:    []atom.Link{atom.Link{Href: inUrl, Rel: "alternate"}},
		Updated: atom.Time(updateTime),
		Author:  &atom.Person{Name: getFormatedName(inResult.Comments[0].Who)},
		Entry:   make([]*atom.Entry, 0, len(inResult.Comments)),
	}

	for i, comment := range inResult.Comments {
		creationTime, err := time.Parse(bugzillaDateFormat, comment.When)
		if err != nil {
			return "", errors.New(fmt.Sprintf("Couldn't parse updateTime in comment %d: %s", i, err))
		}

		// Some xmls do not contain comment_count for whatever reason
		if comment.CommentCount == nil {
			comment.CommentCount = new(int)
			*comment.CommentCount = i
		}

		links := []atom.Link{atom.Link{Href: inUrl + "#c" + strconv.Itoa(*comment.CommentCount), Rel: "alternate"}}
		if comment.AttachmentID != 0 {
			links = append(links, atom.Link{Href: attachmentUrl + strconv.Itoa(comment.AttachmentID), Rel: "enclosure", Title: "Attachment"})
		}

		body := html.EscapeString(comment.Text)
		body = urlRegex.ReplaceAllString(body, `<a href="${all}">${all}</a>`)
		body = attachmentUploadRegex.ReplaceAllString(body, `${pre}<a href="`+attachmentUrl+`${num}">${all}</a>`)
		body = bugNumberRegex.ReplaceAllString(body, `${pre}<a href="`+bugUrl+`${num}">${all}</a>`)
		body = `<pre style="white-space: pre-wrap">` + body + "</pre>"

		// We don't limit the title here too much, because comment.Who can be too
		// large as well. Just to be sure: It needs to be least 3 for the dots.
		maxTitleLength := 100
		title := getFormatedName(comment.Who) + ": " + comment.Text[:min(maxTitleLength, len(comment.Text))]
		if len(title) > maxTitleLength {
			// Find latest space starting three positions before maxTitleLength and cut off
			// If the space is too far away simply cut off without a nice cut.
			// Find lastSpace with maxTitleLength-2, because if this position is a
			// space it is cut out, as title[:lastspace] always cuts at least one char.
			lastSpace := strings.LastIndex(title[:maxTitleLength-2], " ")

			if lastSpace < maxTitleLength*3/4 {
				title = title[:maxTitleLength-3] + "..."
			} else {
				title = title[:lastSpace] + "..."
			}
		}

		entry := &atom.Entry{
			Title:     title,
			ID:        inUrl + "#c" + strconv.Itoa(*comment.CommentCount),
			Link:      links,
			Published: atom.Time(creationTime),
			Updated:   atom.Time(creationTime),
			Author:    &atom.Person{Name: getFormatedName(comment.Who)},
			Content:   &atom.Text{Type: "html", Body: body},
		}

		feed.Entry = append(feed.Entry, entry)
	}

	atom, err := xml.MarshalIndent(feed, "", "\t")
	if err != nil {
		return "", errors.New(fmt.Sprintf("Error during creating the atom feed: %s", err))
	}

	return xml.Header + string(atom), nil
}

func limitStream(fd int, read bool) {
	var rights *unix.CapRights
	rights, err := unix.CapRightsInit(nil)
	if err != nil {
		log.Fatal(err)
	}

	var additionalRights []uint64
	if read {
		additionalRights = []uint64{unix.CAP_READ}
	} else {
		additionalRights = []uint64{unix.CAP_WRITE}
	}

	if err := unix.CapRightsSet(rights, additionalRights); err != nil {
		log.Fatal(err)
	}

	if err := unix.CapRightsLimit(uintptr(fd), rights); err != nil && err != unix.ENOSYS {
		log.Fatal(err)
	}
}

func main() {
	limitStream(unix.Stdin, true)
	limitStream(unix.Stdout, false)
	limitStream(unix.Stderr, false)

	if _, _, err := unix.Syscall(unix.SYS_CLOSEFROM, uintptr(unix.Stderr + 1), 0, 0); err != 0 {
		log.Fatal(err)
	}

	if err := unix.CapEnter(); err != nil && err != unix.ENOSYS {
		log.Fatal(err)
	}

	inXml, err := ioutil.ReadAll(os.Stdin)
	if err != nil {
		log.Fatal(err)
	}

	atom, err := convertXmlToAtom(inXml)
	if err != nil {
		log.Fatal(err)
	}

	os.Stdout.Write([]byte(atom))
}
